{-# LANGUAGE BlockArguments #-}
{-# LANGUAGE DeriveFoldable #-}
{-# LANGUAGE DeriveFunctor #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE StrictData #-}
{-# LANGUAGE UndecidableInstances #-}

-- | A naive implementation of the rtree.
module Control.Monad.RTree (
  -- * RTree
  RTree,
  extract,
  inputs,
  iinputs,
  probe,
  reduce,
  drawRTree,

  -- * RTreeT and RTreeN
  RTreeT (..),
  extractT,
  reduceT,
  RTreeN (..),
  flattenT,

  -- * Re-exports
  module Control.Monad.Reduce,
  module Data.RPath,
) where

import Control.Monad
import Control.Monad.Identity
import Control.Monad.Reduce
import Data.Foldable
import Data.Foldable.WithIndex
import Data.Function ((&))
import Data.RPath
import qualified Data.Sequence as Seq

-- | The simple RTree
data RTree l i
  = Done !i
  | Split !l ~(RTree l i) !(RTree l i)
  deriving (Functor, Foldable)

instance Applicative (RTree l) where
  pure = Done
  (<*>) = ap

instance Monad (RTree l) where
  ma >>= f = case ma of
    Done i -> f i
    Split ctx lhs rhs -> Split ctx (lhs >>= f) (rhs >>= f)

instance MonadReduce l (RTree l) where
  split = Split

instance FoldableWithIndex RPath (RTree l) where
  ifoldMap f =
    Seq.empty & fix \rec rs -> \case
      Done i -> f (fromChoiceList (toList rs)) i
      Split _ lhs rhs -> rec (rs Seq.|> True) lhs <> rec (rs Seq.|> False) rhs

-- | Extract the top value from the RTree.
extract :: RTree l i -> i
extract = \case
  Split _ _ rhs -> extract rhs
  Done i -> i
{-# INLINE extract #-}

-- | A list of inputs, A simple wrapper around @toList@
inputs :: RTree l i -> [i]
inputs = toList

-- | A list of indexed inputs, A simple wrapper around @itoList@
iinputs :: RTree l i -> [(RPath, i)]
iinputs = itoList

probe :: RTree l i -> RPath -> (i, [(Bool, l)])
probe tree pth = go 0 id tree
 where
  go idx res = \case
    Done i -> (i, res [])
    Split l rhs lhs ->
      let b = pth `indexChoice` idx
       in go (idx + 1) (((b, l) :) . res) (if b then rhs else lhs)
{-# INLINE probe #-}

-- | For debugging purposes
drawRTree :: (l -> ShowS) -> (i -> ShowS) -> RTree l i -> String
drawRTree ppl ppi = concat . go id
 where
  go x = \case
    Done i ->
      map
        (\a -> " " <> a <> "\n")
        ( lines
            . ppi i
            . showString "\n"
            . x
            $ ""
        )
    Split ctx2 lhs rhs ->
      let (rh : rhs') = go x rhs
          (lh : lhs') = go (ppl ctx2) lhs
       in fold
            [ ["┳━" <> rh]
            , map ("┃ " <>) rhs'
            , ["┗━" <> lh]
            , map ("  " <>) lhs'
            ]

-- | Reduce the tree
reduce
  :: (Monad m)
  => (i -> m Bool)
  -> RTree l i
  -> m i
reduce p = go
 where
  go = \case
    Done i -> pure i
    Split _ lhs rhs -> do
      p (extract lhs) >>= \case
        True -> go lhs
        False -> go rhs
{-# INLINE reduce #-}

-- | An RTreeT Node
data RTreeN l m i
  = DoneN !i
  | SplitN !l !(RTreeT l m i) !(RTreeT l m i)
  deriving (Functor, Foldable)

newtype RTreeT l m i = RTreeT {unRTreeT :: m (RTreeN l m i)}
  deriving (Functor, Foldable)

instance (Monad m) => Applicative (RTreeT l m) where
  pure = RTreeT . pure . DoneN
  (<*>) = ap

instance (Monad m) => Monad (RTreeT l m) where
  RTreeT ma >>= f = RTreeT $ do
    ma >>= \case
      DoneN i -> unRTreeT (f i)
      SplitN ctx lhs rhs -> pure $ SplitN ctx (lhs >>= f) (rhs >>= f)

instance (Monad m) => MonadReduce l (RTreeT l m) where
  split l lhs rhs = RTreeT (pure $ SplitN l lhs rhs)

-- | Extract a value from an @RTreeT@
extractT :: (Monad m) => RTreeT l m b -> m b
extractT (RTreeT m) = m >>= extractN
{-# INLINE extractT #-}

extractN :: (Monad m) => RTreeN l m b -> m b
extractN = \case
  DoneN i -> pure i
  SplitN _ _ rhs -> extractT rhs
{-# INLINE extractN #-}

flattenT :: RTreeT l Identity i -> RTree l i
flattenT (RTreeT (Identity t)) = case t of
  DoneN i -> Done i
  SplitN ctx lhs rhs -> Split ctx (flattenT lhs) (flattenT rhs)

-- | Reduction in @RTreeT@
reduceT
  :: (Monad m, MonadPlus n)
  => (forall a. m a -> n a)
  -- ^ A function to lift m into n
  -> (i -> n Bool)
  -> RTreeT l m i
  -> n i
reduceT lift_ p = fix \rec x ->
  lift_ (unRTreeT x) >>= \case
    DoneN i -> pure i
    SplitN _ lhs rhs -> do
      lift_ (extractT lhs) >>= p >>= \case
        True -> rec lhs
        False -> rec rhs
{-# INLINE reduceT #-}
